/* ********************************************************
    addeg.cpp

    Author = Hideyuki Harada
    Nihon Univ. Tani Research Group


    My Algorithm for resolving MaxClique problem
    designed for "g++ 3.3.3" with "LEDA 4.5"
   ******************************************************** */


#include<LEDA/graph.h>
#include<LEDA/list.h>
#include<time.h>
#include<iostream.h>
#include<fstream.h>
#include<string.h>
#include<LEDA/d_array.h>

#define P_VERSION "addeg 1.04"

using namespace leda;

graph G;           // undirected graph
list<node> V;
list<edge> E;


list<node> Qmax;   // node set for MaxClique
list<node> Q;

d_array<node, int> TEMP_ADDEG;

int Branch_count;  // number of counted SST branch

int upper_count;

void read_graph(int argc, char *argv[]);   // read filename from command line

void initialize(list<node> *R);  // clear Q and put V to R

void show_result(float T, char *argv[]);  // write result

void expand(list<node> R);  // basic algorithm

void DEG_CLQ(void);

void sort_by_addeg(list<node> *R, d_array<node, int> ADDEG);

void sort_by_degree(list<node> *R);

int cmp_addeg(const node &v, const node &u);

int cmp_degree(const node &v, const node &u);

void set_addeg(list<node> *R, d_array<node, int> *ADDEG);

int upper_bound(list<node> R, int up, d_array<node, int> ADDEG);

void check_degree(node v);

//*************************************************


int main(int argc, char *argv[]){

  float T = used_time();

  if(argc != 2){
    cout << "usage: addeg [file]" << endl;
    exit(1);
  }

  read_graph(argc, argv);

  cout << "resolving " << argv[1] << " ..." << endl;

  // start procedure
  DEG_CLQ();

  show_result(T, argv);

  cout << "done." << endl;

  return 0;

}

void DEG_CLQ(void){

  list<node> R;

  initialize(&R);
  sort_by_degree(&R);

  expand(R);

}

void expand(list<node> R){

  node p, v;
  edge e;
  list<node> Rp, H;
  list<edge> he;
  int up = 1, i;
  d_array<node, int> ADDEG;

  set_addeg(&R, &ADDEG);
  sort_by_addeg(&R, ADDEG);

  while(!R.empty()){
    Branch_count++;

    p = R.head();

    --up = upper_bound(R, up, ADDEG);  // ???
    
    if(Q.size() + ADDEG[R[R[up]]] > Qmax.size()){
      Q.append(p);

      Rp.clear();
      // Rp = intersect(R, G.adj_nodes(p));
 
      forall(v, R){
	if((G.adj_nodes(p)).search(v) != NULL && Q.size() + ADDEG[v] > Qmax.size()){
	  Rp.append(v);
	}else{
	  H.append(v);
	}
      }
     
      forall(v, H){
	G.hide_node(v, he);
      }

      if(!Rp.empty() ){
	expand(Rp);
      }else{
	if(Q.size() > Qmax.size()){  // apply new MaxClique
	  Qmax.clear();
	  forall(v, Q){
	    Qmax.append(v);
	  }
	}
      }
      forall(v, H){
	G.restore_node(v);
      }
      forall(e, he){
	G.restore_edge(e);
      }
      H.clear();
      he.clear();
      Q.remove(p);
    }else{
      return;
    }

    R.remove(p);
  }
  return;

}

void sort_by_addeg(list<node> *R, d_array<node, int>ADDEG){
  
  TEMP_ADDEG = ADDEG;

  (*R).sort(*cmp_addeg, 0);

  return;

}


void sort_by_degree(list<node> *R){
  
  (*R).sort(*cmp_degree, 0);

  return;

}

int cmp_addeg(const node &v, const node &u){

  if(TEMP_ADDEG[v] < TEMP_ADDEG[u]){
    return 1;
  }

  else
    if(TEMP_ADDEG[v] == TEMP_ADDEG[u]){
      return 0;
    }

    else
      return -1;
}


int cmp_degree(const node &v, const node &u){

  if(degree(v) < degree(u)){
    return 1;
  }

  else
    if(degree(v) == degree(u)){
      return 0;
    }

    else
      return -1;
}


void read_graph(int argc, char *argv[]){
  node v;

  G.read(argv[1]);
  G.make_undirected();
  forall_nodes(v, G){
    if(degree(v) == 0)
      G.del_node(v);
  }
  E = G.all_edges();
  V = G.all_nodes();
  
  return;
}

void initialize(list<node> *R){

  Q.clear();
  (*R) = G.all_nodes();
  Qmax.clear();
  Branch_count = 0;
  upper_count =0;

  return;
}

void show_result(float T, char *argv[]){

  node v;
  ofstream fout;
  time_t t=time(NULL);
  string ofile;

  ofile = ".result";
 
  fout.open(argv[1]+ofile, ios::app);

  fout << asctime(localtime(&t));
  fout << "Program version = " << P_VERSION << endl;  
  fout << "Filename = " << argv[1] << endl;
  fout << " Number of vertexes = " << G.number_of_nodes() << endl;
  fout << " Number of edges = " << G.number_of_edges() << endl;
  fout << "Used time = " << used_time(T) << endl;
  fout << "SST Branch count = " << Branch_count << endl;
  fout << "node count for upper bound = " << upper_count << endl;
  fout << "MAX CLIQUE SIZE = " << Qmax.size() << endl;
  fout << "CLIQUE ELEMENT(S) :" << endl;

  forall(v, Qmax){
    fout << " " << index(v);
  }
  fout << endl;

  fout << endl;

  fout.close();
 
  return;
}


void set_addeg(list<node> *R, d_array<node, int> *ADDEG){
  
  node v;
  int max, i;
  list<node> L;

  forall(v, *R){
    L = G.adj_nodes(v);
    sort_by_degree(&L);

    i=0;
    if(!(L.empty())){
      while(i+1 < degree(L[L[i]]) && i < L.size()-1){
        i++;
      }
      if(i < L.size()-1){
	(*ADDEG)[v] = degree(L[L[i]])+1;
      }else{
	//(*ADDEG)[v] = degree(v)+1;
	(*ADDEG)[v] = degree(L[L[L.size()-1]])+1;
      }
    }else{
      (*ADDEG)[v] = 1;
    }
    if((*ADDEG)[v]> degree(v)){
      (*ADDEG)[v] = degree(v)+1;
    }
  }
    /*
    max = degree((*R)[(G.adj_nodes(v)).max(cmp_deg)]);
    if(degree(v) < max){
      ADDEG[v] = max + 1;
    }else{
      ADDEG[v] = degree(v) + 1;
    }
  }
    */

  return;
}

int cmp_deg(const node &v, const node &u){

  return(compare(degree(v), degree(u)));

}


int upper_bound(list<node> R, int up, d_array<node, int> ADDEG){
  
  int i = up;
  
  if(i<0)
    return 0;

  if(!(R.empty()) && i < R.size()-1 ){
    while(i+1 < ADDEG[R[R[i]]] && i < R.size() - 1){
      i++;
    }
    upper_count += (i-up);
  }else{
    return up;
  }
  /*
    if(!(L.empty())){
      while(i+1 < degree(L[L[i]]) && i < L.size()-1){
        i++;
      }
      if(i < L.size()-1){
	(*ADDEG)[v] = degree(L[L[i]])+1;
      }else{
	//(*ADDEG)[v] = degree(v)+1;
	(*ADDEG)[v] = degree(L[L[L.size()-1]])+1;
      }
    }else{
      (*ADDEG)[v] = 1;
    }
    if((*ADDEG)[v]> degree(v)){
    (*ADDEG)[v] = degree(v);
      }
    }
  */

  return i;
}

void check_degree(node v){

  node u;
  list<node> N;
  
  N = G.adj_nodes(v);

  G.hide_node(v);

  forall(u, N){
    if(degree(u) <= Qmax.size()){
      check_degree(u);
    }
  }
  
  return;

}
